/**	AirportMap. A project developep as a part of
 * the Object Oriented Programming in Java
 * @author: Carlos Mario Granados Castro
 * @ date: 08.02.2020
 */

package airportProject;

import java.util.*;

import de.fhpotsdam.unfolding.UnfoldingMap;
import de.fhpotsdam.unfolding.core.Coordinate;
import de.fhpotsdam.unfolding.data.*;
import de.fhpotsdam.unfolding.geo.Location;
import de.fhpotsdam.unfolding.marker.AbstractShapeMarker;
import de.fhpotsdam.unfolding.marker.Marker;
import de.fhpotsdam.unfolding.marker.MultiMarker;
import de.fhpotsdam.unfolding.marker.SimpleLinesMarker;
import de.fhpotsdam.unfolding.providers.*;
import de.fhpotsdam.unfolding.utils.MapUtils;
import de.fhpotsdam.unfolding.data.Feature;
import de.fhpotsdam.unfolding.data.PointFeature;
import de.fhpotsdam.unfolding.data.ShapeFeature;

import parsing.ParseFeed;
import processing.core.PApplet;

public class AirportMap extends PApplet {

	// The map
	private UnfoldingMap map;
	
	// The files containing city names and info and country names and info
	private String cityFile = "city-data.json";
	private String countryFile = "countries.geo.json";
	
	// The files containing airport info and routes
	private String airportFile = "airports.dat";
	private String routeFile = "routes.dat";
	
	// The markers for the cities
	private List<Marker> cityMarkers;
	
	// Markers for airports and routes
	private List<Marker> airportMarkers;
	private List<Marker> routeMarkers;
	private List<ShapeFeature> routes;
	
	// A List of country markers
	private List<Marker> countryMarkers;
		
	// To watch mouse actions
	private CommonMarker lastSelected;
	private CommonMarker lastClicked;
		
	// Custom colors, based on Google Material Design
	protected int yellow = color(255, 202, 40);
	protected int red = color(229, 57, 53);
	protected int blue = color(13, 71, 161);
	protected int purple = color(94, 53, 177);
	protected int grey = color(189, 189, 189);
	protected int blueLight = color(41, 182, 246);
	protected int blueDark = color(13, 71, 161);
	
	// Distances to consider airports...
	private double distMax = 200;
	private double distMin = 75;
	protected static float altMin;
	protected static float altMax;
	
	// Method to initialize and setup canvas, map, markers...
	public void setup() {
		
		// Initialize canvas
		size(900, 700, OPENGL);
		
		// To select map providers
		//AbstractMapProvider provider = new AcetateProvider.Terrain();	//Not working?
		//AbstractMapProvider provider = new EsriProvider.NatGeoWorldMap();
		AbstractMapProvider provider = new EsriProvider.WorldPhysical();
		//AbstractMapProvider provider = new EsriProvider.WorldTerrain();
		//AbstractMapProvider provider = new Google.GoogleMapProvider();
		//AbstractMapProvider provider = new Google.GoogleTerrainProvider();
		//AbstractMapProvider provider = new ImmoScout.HeatMapProvider(); //Not working?
		//AbstractMapProvider provider = new MapBox.WorldLightProvider(); //Not working?
		//AbstractMapProvider provider = new MapQuestProvider.Aerial(); //Not working?
		//AbstractMapProvider provider = new Microsoft.AerialProvider();
		//AbstractMapProvider provider = new Microsoft.RoadProvider();
		//AbstractMapProvider provider = new OpenStreetMap.OpenStreetMapProvider();
		//AbstractMapProvider provider = new OpenWeatherProvider.Clouds(); //Not working?
		//AbstractMapProvider provider = new OpenWeatherProvider.Temperature(); //Not working?
		//AbstractMapProvider provider = new ThunderforestProvider.Landscape();
		//AbstractMapProvider provider = new Yahoo.AerialProvider(); //Not working?
		//AbstractMapProvider provider = new Yahoo.RoadProvider(); //Not working?
				
		// Create map and setup interactive options
		map = new UnfoldingMap(this, 200, 50, 650, 600, provider);
		MapUtils.createDefaultEventDispatcher(this, map);
		
		// Read country and city features and data
		List<Feature> countries = GeoJSONReader.loadData(this, countryFile);
		List<Feature> cities = GeoJSONReader.loadData(this, cityFile);
		countryMarkers = MapUtils.createSimpleMarkers(countries);
		cityMarkers = new ArrayList<Marker>();
		for (Feature city : cities) {
			cityMarkers.add(new CityMarker(city));
		}
	
		// Read airport data (features) and create markers
		List<PointFeature> features = parseAirports(this, airportFile);
		airportMarkers = new ArrayList<Marker> ();
		HashMap<Integer, Location> airports = new HashMap<Integer, Location>();
		for (PointFeature feature : features) {
			AirportMarker m = new AirportMarker(feature);
			airportMarkers.add(m);
			airports.put(Integer.parseInt(feature.getId()), feature.getLocation());
		}
		
		checkAirportsCities(distMax);
		
		// Read route data
		List<ShapeFeature> routes = parseRoutes(this, routeFile);
		routeMarkers = new ArrayList<Marker>();
		for (ShapeFeature route : routes) {
			// Get source and destination airportIDs
			int source = Integer.parseInt((String)route.getProperty("source"));
			int dest = Integer.parseInt((String)route.getProperty("destination"));
			
			// Get locations for airports on route
			if (airports.containsKey(source) && airports.containsKey(dest)) {
				route.addLocation(airports.get(source));
				route.addLocation(airports.get(dest));
			}
			
			SimpleLinesMarker sl = new SimpleLinesMarker(route.getLocations(), route.getProperties());
			
			//System.out.println(sl.getProperties());
			
			// To test routeMarkers
			sl.setHidden(true);
			routeMarkers.add(sl);
			
		}
		
		// Add markers to map
		map.addMarkers(cityMarkers);
		map.addMarkers(airportMarkers);
		map.addMarkers(routeMarkers);
		
	}
	
	// Method to draw map and conventions...
	public void draw() {
 	
		background(0);
		map.draw();
		addKey();
		
	}
	
	// Event handler that gets called automatically when mouse moves
	@Override
	public void mouseMoved() {
		// Clear last selection
		if (lastSelected != null) {
			lastSelected.setSelected(false);
			lastSelected = null;
		}
		selectMarkerIfHover(cityMarkers);
		selectMarkerIfHover(airportMarkers);
	}
	
	// If there is a marker selected...
	private void selectMarkerIfHover(List<Marker> markers) {
		
		// Abort if there's already a marker selected
		if (lastSelected != null) {
			return;
		}
		
		for (Marker m : markers) {
			CommonMarker marker = (CommonMarker)m;
			if (marker.isInside(map, mouseX, mouseY)) {
				lastSelected = marker;
				marker.setSelected(true);
				return;
			}
		}
	}
	
	// Helper method, to hide all airports that are farther than distRef for each city
	//	in cityMarkers
	private void checkAirportsCities(double distRef) {
		
		for (Marker mhide : airportMarkers) {
			mhide.setHidden(true);
			int indxCond = 1;
			int indx = 0;
			while (indxCond == 1) {
			    Marker m = cityMarkers.get(indx);
				double dist = mhide.getDistanceTo(m.getLocation());
				if (dist < distRef) {
					mhide.setHidden(false);
					indxCond = 0;
				}
				indx ++;
				if (indx >= cityMarkers.size()) {
					indxCond = 0;
				}
			}
		}
	
	}

	// Helper method, to hide all airports that are farther than distRef for a given city
	private void checkAirportsCities(Marker city, double distRef) {
		
		for (Marker mhide : airportMarkers) {
			mhide.setHidden(true);
			double dist = mhide.getDistanceTo(city.getLocation());
			if (dist < distRef) {
				mhide.setHidden(false);
			}
		}
	
	}
	
	// Event handler for mouse clicks
	@Override
	public void mouseClicked() {
		
		if (lastClicked != null) {
			unhideMarkers();
			lastClicked = null;
		}
		else {
			checkCitiesForClick();
			if (lastClicked == null) {
				checkAirportsForClick();
			}
		}
		
	}
	
	private void unhideMarkers() {
		for (Marker m : cityMarkers) {
			m.setHidden(false);
		}
		for (Marker sl : routeMarkers) {
			sl.setHidden(true);
		}
		checkAirportsCities(distMax);
	}
	
	// Helper method to determine actions when the user clicks over a city
	private void checkCitiesForClick() {
		
		if (lastClicked != null ) {
			return;
		}
		
		// Loop over cities to see if one of them is selected
		for(Marker m : cityMarkers) {
			if (!m.isHidden() && m.isInside(map, mouseX, mouseY)) {
				lastClicked = (CommonMarker)m;
				checkAirportsCities(m, distMin);
			}
		}
		
	}
	
	// Helper method to determine actions when the user clicks over an airport
	private void checkAirportsForClick() {
		
		if (lastClicked != null) {
			return;
		}
		
		ArrayList<Integer> destList = new ArrayList<Integer>();
		// Loop over airports to see if one of them is selected
		for (Marker m : airportMarkers) {
			if (!m.isHidden() && m.isInside(map, mouseX, mouseY)) {
				lastClicked = (CommonMarker)m;
				AirportMarker airMarker = (AirportMarker)m;
				int sourceRef = airMarker.getID();
				destList.add(sourceRef);
				// To unhide all the routes that has as source or destination the selected airport.
				for (Marker sl : routeMarkers) {
					int source = Integer.parseInt((String)sl.getProperty("source"));
					int dest = Integer.parseInt((String)sl.getProperty("destination"));
					if (source == sourceRef || dest == sourceRef) {
						sl.setHidden(false);
						if (!destList.contains(source)) {
							destList.add(source);
						}
						if (!destList.contains(dest)) {
							destList.add(dest);
						}
					}
				}
			}
		}
		
		for (Marker m : airportMarkers) {
			
			AirportMarker airMarker = (AirportMarker)m;
			int markID = airMarker.getID();
			if (destList.contains(markID)) {
				m.setHidden(false);
			}
			else {
				m.setHidden(true);
			}
		}
			
	}
	
	private void addKey() {
		
		fill(grey);
		
		int x0 = 25;
		int y0 = 50;
		
		int dx = 25;
		int dy = 35;
		
		rect(x0, y0, 150, 250);
		
		// Title
		y0 += 25;
		fill(0);
		textAlign(LEFT, CENTER);
		textSize(12);
		text("Map Keys", x0+dx, y0);
		
		// Forms
		fill(purple);
		int xbase = x0 + dx;
		int ybase = y0 + dy;
		int dt = CityMarker.SIZE;
		triangle(xbase, ybase-dt, xbase-dt, ybase+dt, xbase+dt, ybase+dt);
		fill(blueLight);
		ellipse(xbase, ybase+dy, 10, 10);
		
		// Text
		fill(0);
		text("City Marker", xbase + dx, ybase);
		text("Airport Marker", xbase + dx, ybase+dy);
		
		
	}
	
	/*
	 * This method is to parse a file containing airport information.  
	 * The file and its format can be found: 
	 * http://openflights.org/data.html#airport
	 * 
	 * It is also included with the UC San Diego MOOC package in the file airports.dat
	 * 
	 * @param p - PApplet being used
	 * @param fileName - file name or URL for data source
	 */
	public List<PointFeature> parseAirports(PApplet p, String fileName) {
		List<PointFeature> features = new ArrayList<PointFeature>();

		altMin = 0;
		altMax = 0;
		
		String[] rows = p.loadStrings(fileName);
		for (String row : rows) {
			
			// hot-fix for altitude when lat lon out of place
			int i = 0;
			
			// split row by commas not in quotations
			String[] columns = row.split(",(?=([^\"]*\"[^\"]*\")*[^\"]*$)");
			
			// get location and create feature
			//System.out.println(columns[6]);
			float lat = Float.parseFloat(columns[6]);
			float lon = Float.parseFloat(columns[7]);
			
			Location loc = new Location(lat, lon);
			PointFeature point = new PointFeature(loc);
			
			// set ID to OpenFlights unique identifier
			point.setId(columns[0]);
			
			// get other fields from csv
			point.addProperty("ID", columns[0]);
			point.addProperty("name", columns[1]);
			point.putProperty("city", columns[2]);
			point.putProperty("country", columns[3]);
			
			// pretty sure IATA/FAA is used in routes.dat
			// get airport IATA/FAA code
			if(!columns[4].equals("")) {
				point.putProperty("code3", columns[4]);
			}
			// get airport ICAO code if no IATA
			else if(!columns[5].equals("")) {
				point.putProperty("code4", columns[5]);
			}
			
			point.putProperty("altitude", columns[8 + i]);
			float alt = Float.parseFloat(columns[8+i]);
			if(altMin == 0.0) {
				altMin = alt;
			}
			if(alt < altMin) {
				altMin = alt;
			}
			if(alt > altMax) {
				altMax = alt;
			}
			
			features.add(point);
		}

		return features;
		
	}
	
	/*
	 * This method is to parse a file containing airport route information.  
	 * The file and its format can be found: 
	 * http://openflights.org/data.html#route
	 * 
	 * It is also included with the UC San Diego MOOC package in the file routes.dat
	 * 
	 * @param p - PApplet being used
	 * @param fileName - file name or URL for data source
	 */
	public static List<ShapeFeature> parseRoutes(PApplet p, String fileName) {
		List<ShapeFeature> routes = new ArrayList<ShapeFeature>();
		
		String[] rows = p.loadStrings(fileName);
		
		for(String row : rows) {
			String[] columns = row.split(",");
			
			ShapeFeature route = new ShapeFeature(Feature.FeatureType.LINES);
			
			// set id to be OpenFlights identifier for source airport
			
			// check that both airports on route have OpenFlights Identifier
			if(!columns[3].equals("\\N") && !columns[5].equals("\\N")){
				// set "source" property to be OpenFlights identifier for source airport
				route.putProperty("source", columns[3]);
				// "destination property" -- OpenFlights identifier
				route.putProperty("destination", columns[5]);
				
				routes.add(route);
			}
		}
			
		
		return routes;
		
	}
	
}
